# =============================================================================
# Ural Redirecion Inferrence Function
# =============================================================================
#
# A lot of urls contains an obvious hint that they will in fact trigger
# a redirection. This modules gathers routines aimed at discovering
# those redirections without even firing a HTTP request.
#
import re

from ural.patterns import QUERY_VALUE_IN_URL_TEMPLATE
from ural.utils import unquote, urljoin

OBVIOUS_REDIRECTS_RE = re.compile(
    QUERY_VALUE_IN_URL_TEMPLATE
    % r"(?:redirect(?:_to)?|target|redir|next|link|orig|goto|url|[luq])",
    re.I,
)
REDIRECTION_DOMAINS_RE = re.compile(
    r"(?:\.ampproject\.org/[cv]/(?:s/)?|bc\.marfeelcache\.com/amp/|bc\.marfeel\.com/)",
    re.I,
)


def infer_redirection(url, recursive=True):
    """
    Function returning the url that the given url will redirect to. This is done
    by finding obvious hints in the GET parameters that the given url is in
    fact a redirection.

    Args:
        url (string): Target url.
        recursive (bool): Whether to apply the function recursively until
            no redirection can be inferred. Defaults to `True`.

    Returns:
        string: Redirected url or the original url if nothing was found.
    """

    redirection_split = REDIRECTION_DOMAINS_RE.split(url, 1)

    target = None

    if len(redirection_split) > 1:
        # NOTE: avoiding empty AMP redirects etc.
        if len(redirection_split[1]):
            target = "https://" + redirection_split[1]

    else:
        obvious_redirect_match = re.search(OBVIOUS_REDIRECTS_RE, url)

        if obvious_redirect_match is not None:
            potential_target = unquote(obvious_redirect_match.group(1))

            # NOTE: the len check is here to fend off empty redirects
            if potential_target.startswith("https://") and len(potential_target) > 8:
                target = potential_target
            elif potential_target.startswith("http://") and len(potential_target) > 7:
                target = potential_target

            if potential_target.startswith("/"):
                target = urljoin(url, potential_target)

    if target is None:
        return url

    if recursive:
        return infer_redirection(target, recursive=True)

    return target



# -*- coding: utf-8 -*-
# =============================================================================
# Ural URL Is Typo Url Function
# =============================================================================
#
# A function returning True if its argument is detected as typo.
#
from __future__ import unicode_literals
from ural.strip_protocol import strip_protocol
from ural.utils import safe_urlsplit

ERROR_TLDS = {
    "cab",
    "global",
    "ren",
    "gay",
    "baby",
    "gallery",
    "red",
    "tattoo",
    "lincoln",
    "ooo",
    "new",
    "barcelona",
    "med",
    "photos",
    "africa",
    "film",
    "sale",
    "amazon",
    "rip",
    "love",
    "py",
    "android",
    "video",
    "kim",
    "ro",
    "ck",
    "ba",
    "day",
    "bayern",
    "mm",
    "luxe",
    "blog",
    "mo",
    "gm",
    "design",
    "vote",
    "pub",
    "blue",
    "bn",
    "zip",
    "style",
    "nu",
    "ml",
    "mil",
    "youtube",
    "eco",
    "place",
    "et",
    "fk",
    "ni",
    "rent",
    "bond",
    "prod",
    "post",
    "berlin",
    "protection",
    "you",
    "gh",
    "by",
    "clothing",
    "ls",
    "np",
    "sexy",
    "il",
    "ferrero",
    "map",
    "bom",
    "bb",
    "skin",
    "bible",
    "buy",
    "market",
    "ice",
    "cfa",
    "group",
    "office",
    "meme",
    "dm",
    "vi",
    "xxx",
    "audio",
    "dupont",
    "cv",
    "science",
    "pharmacy",
    "systems",
    "ga",
    "contact",
    "family",
    "sport",
    "gallup",
    "zara",
    "ms",
    "cz",
    "sj",
    "maison",
    "dj",
    "hockey",
    "pk",
    "cash",
    "ag",
    "cr",
    "ink",
    "website",
    "mt",
    "pr",
    "bio",
    "ses",
    "lol",
    "iq",
    "norton",
    "kh",
    "la",
    "hair",
    "chat",
    "wow",
    "bo",
    "mg",
    "ss",
    "osaka",
    "sh",
    "ge",
    "lifestyle",
    "now",
    "tel",
    "mr",
    "click",
    "ping",
    "earth",
    "eus",
    "shell",
    "football",
    "tours",
    "surgery",
    "dot",
    "mma",
    "zero",
    "ing",
    "read",
    "credit",
    "mc",
    "black",
    "inc",
    "tennis",
    "author",
    "like",
    "social",
    "mov",
    "er",
    "apple",
    "moi",
    "google",
    "suzuki",
    "smile",
    "able",
    "xin",
    "sm",
    "sx",
    "plus",
    "ao",
    "am",
    "lk",
    "no",
    "box",
    "vision",
    "gp",
    "vu",
    "name",
    "delta",
    "ve",
    "quebec",
    "lat",
    "off",
    "house",
    "ltd",
    "na",
    "bs",
    "bmw",
    "game",
    "rugby",
    "mw",
    "total",
    "aw",
    "man",
    "ceo",
    "madrid",
    "star",
    "build",
    "data",
    "ye",
    "pictures",
    "gu",
    "gle",
    "baseball",
    "je",
    "car",
    "helsinki",
    "buzz",
    "im",
    "kn",
    "mu",
    "sr",
    "va",
    "pa",
    "sakura",
    "fox",
    "phone",
    "sa",
    "ki",
    "tiffany",
    "ps",
    "rocher",
    "re",
    "men",
    "beauty",
    "boo",
    "sy",
    "pet",
    "km",
    "gb",
    "open",
    "pin",
    "jo",
    "zw",
    "ne",
    "so",
    "ky",
    "play",
    "pics",
    "mtn",
    "dvr",
}

ERROR_INCLUSIVE = [
    "é.es",
    "eux.se",
    "s.es",
    "eur.se",
    "eu.r.se",
    "ant.es",
    "eu.x.se",
    "eu.se",
    "un.es",
    ".e.es",
    "eux.ses",
    "r.es",
    "ois.es",
    "t.es",
    "l.es",
    "i.es",
    "n.es",
    "u.es",
]


def is_typo_url(url):
    """
    Function returning True if its argument is detected as url typo.

    Args:
        url (str): string to test.

    Returns:
        bool: True if the argument contains typo, False if not.

    """
    # tests if there is a '/' in the url except if it is at the end
    protocoleless_url = strip_protocol(url)
    slash_count = protocoleless_url.count("/")

    if slash_count > 1:
        return False
    elif slash_count == 1 and not protocoleless_url.endswith("/"):
        return False

    # tests if there is a ? in the url
    if "?" in protocoleless_url:
        return False

    # tests if '.co' is contained in the url
    if ".co." in protocoleless_url:
        return False

    # tests if there is a # in the url
    if "#" in protocoleless_url:
        return False

    # tests if domain starts with 'www'
    domain = safe_urlsplit(url).hostname
    if domain.startswith("www"):
        return False

    url = url.rstrip("/")
    _, tld = url.rsplit(".", 1)

    return (
        (any(char.islower() for char in tld) and any(char.isupper() for char in tld))
        or tld.lower() in ERROR_TLDS
        or is_inclusive_language(url)
    )


def is_inclusive_language(string):
    """
    Function returning True if its argument contains inclusive language (in french)
    """
    for pattern in ERROR_INCLUSIVE:
        if string.lower().endswith(pattern):
            return True

    return False



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Has Special Host Function
# =============================================================================
#
# Function returning whether the given url looks like it has a special host.
#

from ural.utils import safe_urlsplit
from ural.patterns import SPECIAL_HOSTS_RE


def is_special_host(hostname):
    """
    Function returning whether the given hostname looks like a special host.

    Args:
        hostname (str): hostname to test.

    Returns:
        bool

    """
    return bool(SPECIAL_HOSTS_RE.match(hostname))


def has_special_host(url):
    """
    Function returning whether the given url looks like it has a special host.

    Args:
        url (str): url to test.

    Returns:
        bool

    """
    parsed = safe_urlsplit(url)

    return is_special_host(parsed.hostname)



from __future__ import print_function

from argparse import ArgumentParser

parser = ArgumentParser(prog="ural")
subparsers = parser.add_subparsers(dest="action")
_ = subparsers.add_parser(
    "upgrade",
    help="Run this command to upgrade ural's TLD utilities wrt latest Mozilla data.",
)

cli_args = parser.parse_args()

if cli_args.action == "upgrade":
    from ural.tld import upgrade

    upgrade()



# =============================================================================
# TrieDict
# =============================================================================
#
# A simple Python implementation of a (prefix, value) Trie.
#
NULL = object()


class TrieDictNode(object):
    __slots__ = ("children", "value", "counter")

    def __init__(self):
        self.children = None
        self.value = NULL
        # Counts the number of items in the node's children
        self.counter = 0


class TrieDict(object):
    __slots__ = "__root"

    def __init__(self):
        self.__root = TrieDictNode()

    def __len__(self):
        return self.__root.counter

    def __setitem__(self, prefix, value):
        node = self.__root

        visited_nodes = []

        for token in prefix:

            if node.children is None:
                child = TrieDictNode()

                node.children = {token: child}

                visited_nodes.append(node)
                node = child
                continue

            child = node.children.get(token)

            if child is not None:
                visited_nodes.append(node)
                node = child
            else:
                child = TrieDictNode()

                node.children[token] = child
                visited_nodes.append(node)
                node = child

        if node.value is NULL:
            for n in visited_nodes:
                n.counter += 1

        node.value = value

    def get(self, prefix, default=None):
        node = self.__root

        for token in prefix:
            if node.children is None:
                return default

            child = node.children.get(token)

            if child is None:
                return default

            node = child

        if node.value is not NULL:
            return node.value

        return default

    def __getitem__(self, prefix):
        node = self.__root

        for token in prefix:
            if node.children is None:
                raise KeyError(prefix)

            child = node.children.get(token)

            if child is None:
                raise KeyError(prefix)

            node = child

        if node.value is not NULL:
            return node.value

        raise KeyError(prefix)

    def longest_matching_prefix_value(self, prefix):
        node = self.__root

        last_value = NULL

        for token in prefix:

            if node.value is not NULL:
                last_value = node.value

            if node.children is None:
                break

            child = node.children.get(token)

            if child is None:
                break

            node = child

        if node.value is not NULL:
            return node.value

        return last_value if last_value is not NULL else None

    def items(self):
        stack = [(self.__root, [])]

        while len(stack) > 0:
            node, prefix = stack.pop()

            if node.value is not NULL:
                yield (prefix, node.value)

            if node.children is None:
                continue

            for token, child in node.children.items():
                stack.append((child, prefix + [token]))

    def prefixes(self):
        stack = [(self.__root, [])]

        while len(stack) > 0:
            node, prefix = stack.pop()

            if node.value is not NULL:
                yield prefix

            if node.children is None:
                continue

            for token, child in node.children.items():
                stack.append((child, prefix + [token]))

    def values(self):
        stack = [self.__root]

        while len(stack) > 0:
            node = stack.pop()

            if node.value is not NULL:
                yield node.value

            if node.children is None:
                continue

            stack.extend(node.children.values())

    def __iter__(self):
        return self.items()

    def set_and_prune_if_shorter(self, prefix, value):

        node = self.__root

        visited_nodes = []

        for token in prefix:

            # Check if we try to add a longer prefix
            if node.value is not NULL:
                return

            if node.children is None:
                child = TrieDictNode()

                node.children = {token: child}
                visited_nodes.append(node)
                node = child
                continue

            child = node.children.get(token)

            if child is not None:
                visited_nodes.append(node)
                node = child
            else:
                child = TrieDictNode()

                node.children[token] = child
                visited_nodes.append(node)
                node = child

        # Trie already has longer prefixes of the prefix we are trying to add : we delete those prefixes and new (shortest) one
        if node.children is not None:
            node.children = None
            for n in visited_nodes:
                n.counter -= node.counter - 1
            node.counter = 0

        # We add a prefix with no longer prefixes in the Trie
        elif node.value is NULL:
            for n in visited_nodes:
                n.counter += 1

        node.value = value



# =============================================================================
# Ural Telegram-related heuristic functions
# =============================================================================
#
# Collection of functions related to Telegram urls.
#
import re
from collections import namedtuple

from ural.ensure_protocol import ensure_protocol
from ural.utils import pathsplit, urlsplit, urlunsplit, safe_urlsplit, SplitResult
from ural.patterns import DOMAIN_TEMPLATE

TELEGRAM_MESSAGE_ID_RE = re.compile(r"^\d+$")
TELEGRAM_DOMAINS_RE = re.compile(r"(?:telegram\.(?:org|me)|t\.me)$", re.I)
TELEGRAM_URL_RE = re.compile(
    DOMAIN_TEMPLATE % r"(?:[^.]+\.)*(?:telegram\.(?:org|me)|t\.me)", re.I
)
TELEGRAM_PUBLIC_REPLACE_RE = re.compile(
    r"^(?:[^.]+\.)?(?:telegram\.(?:org|me)|t\.me)", re.I
)

TelegramMessage = namedtuple("TelegramMessage", ["name", "id"])
TelegramGroup = namedtuple("TelegramGroup", ["id"])
TelegramChannel = namedtuple("TelegramChannel", ["name"])


def is_telegram_message_id(value):
    return bool(re.search(TELEGRAM_MESSAGE_ID_RE, value))


def is_telegram_url(url):
    """
    Function returning whether the given url is a valid Telegram url.

    Args:
        url (str): Url to test.

    Returns:
        bool: Whether given url is from Telegram.

    """
    if isinstance(url, SplitResult):
        return bool(re.search(TELEGRAM_DOMAINS_RE, url.hostname))

    return bool(re.match(TELEGRAM_URL_RE, url))


def convert_telegram_url_to_public(url):
    """
    Function parsing the given telegram url and returning the same but the
    the public version.
    """
    safe_url = ensure_protocol(url)

    has_protocol = safe_url == url

    scheme, netloc, path, query, fragment = urlsplit(safe_url)

    if not is_telegram_url(netloc):
        raise TypeError(
            "ural.telegram.convert_telegram_url_to_public: %s is not a telegram url"
            % url
        )

    netloc = re.sub(TELEGRAM_PUBLIC_REPLACE_RE, "t.me/s", netloc)

    result = (scheme, netloc, path, query, fragment)

    result = urlunsplit(result)

    if not has_protocol:
        result = result.split("://", 1)[-1]

    return result


def parse_telegram_url(url):
    """
    Function parsing the given url and returning either a TelegramMessage,
    TelegramChannel, TelegramGroup or None if nothing of information could be
    found.

    Args:
        url (str): Url to parse.

    """
    if not is_telegram_url(url):
        return None

    parsed = safe_urlsplit(url)
    path = pathsplit(parsed.path)

    if path:

        if path[0] == "s":

            if path[1] == "joinchat":
                if len(path) == 3:
                    return TelegramGroup(id=path[2])
                else:
                    None

            elif len(path) == 3 and is_telegram_message_id(path[2]):
                return TelegramMessage(name=path[1], id=path[2])

            elif len(path) == 2:
                return TelegramChannel(name=path[1])

            return None

        else:

            if path[0] == "joinchat":

                if len(path) == 3:
                    return TelegramGroup(id=path[2])

                elif len(path) == 2:
                    return TelegramGroup(id=path[1])

                else:
                    None

            elif len(path) == 2 and is_telegram_message_id(path[1]):
                return TelegramMessage(name=path[0], id=path[1])

            elif len(path) == 1:
                return TelegramChannel(name=path[0])

    return None


def extract_channel_name_from_telegram_url(url):
    parsed = parse_telegram_url(url)

    if parsed is None or isinstance(parsed, TelegramGroup):
        return

    return parsed.name
